import numpy as np
from .jointanimation import JointAnimation

class Animation:
    """Class represent an aimation.

    Contains a list of all joint animations for a given animation.

    Attributes:
    ----------
    duration : float
        The duration of the animation in seconds.
    joint_animations : (JointAnimation) list
        The joint animations that make up the animation.
    """

    def __init__(self, duration = None, joint_animations = []):
        """**Default constructor for the Animation class.**

        Note: If no duration is given, it is computed from joint animations.

        Warns:
        ----------
        UserWarning
            If joint animations duration are not identical.
        
        Parameters:
        ----------
        duration : float, optional
            The duration of the animation in seconds, by default None.
        joint_animations : (JointAnimation) list, optional
            The joint animations that make up the animation, by default empty list.
        """
        self.joint_animations = joint_animations
        self.duration = duration
        if duration is None:
            self.compute_duration()

    def compute_duration(self):
        """Computes the duration of the animation by taking the longest joint animation duration.

        """
        self.duration = max([joint_animation.duration for joint_animation in self.joint_animations])
    
    def as_joint_animation_list(self, root_joint):
        """Returns the animation as a joint animation lists.

        Note: If a joint is present in the hierarchy, but it is not associated with any joint animation, 
            a new JointAnimation is created for the joint with an empty list of keys.

        Returns:
        ----------
        (JointAnimation) list
            List of joint animations in the animation.
        """

        # TODO: move this to utils to avoid redundancy.
        adjacency_list = []

        def dfs(joint):
            adjacency_list.append(joint)
            for child in joint.children:
                dfs(child)

        dfs(root_joint)
        
        joint_animation_list = []
        
        # TODO: improve performance.
        for joint in adjacency_list:
            ja = [ja for ja in self.joint_animations if ja.joint.id == joint.id]
            if ja:
                joint_animation_list.append(ja[0])
            else:
                joint_animation_list.append(JointAnimation(joint, duration = self.duration))
        
        return joint_animation_list
            
     
        

