from .animation import Animation
from ..model.skeleton import Skeleton
from ..math.transform import Transform

class Animator:
    """Controller class to animate a skeleton.

    Attributes:
    ----------
    current_anim : Animation
        Animation to be played.
    current_skeleton : Skeleton
        Skeleton to be animated.
    anim_time : float
        Time elapsed since the beginning of the animation.
    fps : int
        Frames per second of the animation. Corresponds to the number of poses calculated per second.
    interpolation_algo : str, optional
        Algorithm to be used to interpolate between the keys of the animation.
    """

    def __init__(self, current_anim, current_skeleton, fps = 60, interpolation_algo = "Linear"):
        """**Default constructor for the Animator class.**

        Parameters:
        ----------
        current_anim : Animation
            Animation to be played.
        current_skeleton : Skeleton
            Skeleton to be animated.    
        fps : int, optional
            Frames per second of the animation, by default 60.
        interpolation_algo : str, optional
            Interpolation algorithm to be used, by default "Constant".
        """
        self.current_anim = current_anim 
        self.current_skeleton = current_skeleton
        self.anim_time = 0
        self.fps = fps
        self.interpolation_algo = interpolation_algo

    def generate_pose(self, temp_time = 0):
        """Generates a pose.

        Create variable t that is self.anim_time modulo self.current_anim.duration.
        Initialize a new Pose object with the current skeleton.       

        """
        real_time = self.anim_time % self.current_anim.duration
        pose_dict = {}


        for joint_animation in self.current_anim.as_joint_animation_list(self.current_skeleton.root):
            
            joint = joint_animation.joint
            
            if joint.name == "right_arm":
                print("right_arm")
            
            animation_transform = None

            if joint_animation.keys:

                key_time = joint_animation.compute_key_time(real_time)           

                if self.interpolation_algo == "Constant":
                    animation_transform = joint_animation.get_previous_key_transform(key_time)
                elif self.interpolation_algo == "Linear":
                    animation_transform = joint_animation.get_lerp_transform(key_time)
                elif self.interpolation_algo == "Spline":
                    #Not implemented yet
                    pass
            
            else:
                animation_transform = Transform()

            pose_dict[joint.id] = animation_transform

        self.current_skeleton.compute_pose(pose_dict)
        

    def update(self, time):
        """Updates the animation time.

        """
        print("_dt:", self.anim_time)
        self.anim_time += time
    
    def play(self, length = 10, verbose=False):
        """Plays the animation.

        Parameters:
        ----------
        play_duration : int, optional
            The length of time the animation is left runnig in seconds, by default 10.
            
        """
        while self.anim_time <= length:
            self.step(verbose)

    def step(self, time, verbose=False):
        """Steps the animation.

        """
        self.update(time)
        self.generate_pose()
        if verbose:
            print("_Time:", self.anim_time)
            for joint in self.current_skeleton.as_joint_list():
                print("Joint:", joint.name, "Transform:", joint.animated_transform)
        
        

