import warnings
import numpy as np
import math
from .key import Key
from ..math import Transform

class JointAnimation:
    """Class to store a joint's animation data.

    Attributes:
    ----------
    joint : Joint
        The joint that is animated.
    duration : float
        The duration of the joint animation in seconds.
    keys : (Key) list
        The keys of the joint animation.
    time_keys : (float, float) list
        The time keys of the joint animation. They make up the *timewarp curve* 
        which defines the unfolding in time of the sequence of keys for the joint animation.
        A time_keys is a tuple of two floats:
        - The **first value** is the real time *in seconds* between 0 and the duration of the joint animation.
        - The **second value** is the time relative to the keys *between 0 and 1*.
        TODO: explain timewarp curve better.
    """
    def __init__(self, joint, keys = [], time_keys = [], duration = 1):
        """**Default constructor for the JointAnimation class.**

        Parameters:
        ----------
        joint : Joint
            The joint that is animated.
        keys : (Key) list, optional
            The keys of the joint animation, by default empty list.
        time_keys : (float, float) list, optional
            The time keys of the joint animation, by default empty list.
        """
        self.joint = joint
        self.duration = duration
        self.keys = []
        for key in keys:
            self.add_key(key)
        self.time_keys = []
        for time_key in time_keys:
            self.add_time_key(time_key[0], time_key[1])

    def add_key(self, key):
        """Adds a key to the joint animation.

        Raises:
        ----------
        ValueError
            If a key with the same time already exists in joint animation.
        
        Parameters:
        ----------
        key : Key
            The key to be added to the joint animation.
        """
        if self.get_key_by_time(key.time) is None:
            self.keys.append(key)
        else:
            raise ValueError("Key with identical time already in joint animation.")       

    def get_key_by_time(self, time):
        """Returns the key from the joint animation with the given time.

        Parameters:
        ----------
        time : float     
            The time of the key to be returned.

        Returns:
        ----------
        Key
            The key at the given time. None if no key at the given time is found.
        """
        for key in self.keys:
            if math.isclose(key.time, time, rel_tol=1e-09, abs_tol=1e-10):
                return key
        return None

    def add_time_key(self, real_time, key_time):
        """Adds a time key to the joint animation.

        Raises:
        ----------
        ValueError
            If the real time is not between 0 and the duration of the joint animation.
        ValueError
            If a time key with the same real time already exists in joint animation.

        Parameters:
        ----------
        real_time : float
            The real time of the key.
        key_time : float
            The time associated with the key.
        """
        if real_time < 0 or real_time > self.duration:
            raise ValueError("Real time must be between 0 and duration.")
        else:
            if self.get_time_key_by_time(real_time) is None:
                self.time_keys.append((real_time, key_time))
            else:
                raise ValueError("Time key with identical real time already in joint animation.")
        
    
    def get_time_key_by_time(self, real_time):
        """Returns the time key from the joint animation with the given real time.

        Parameters:
        ----------
        real_time : float     
            The real time of the time key to be returned.

        Returns:
        ----------
        (float, float)
            The time key at the given real time. None if no time key at the given real time is found.
        """
        for time_key in self.time_keys:
            if math.isclose(time_key[0], real_time, rel_tol=1e-09, abs_tol=1e-10):
                return time_key
        return None

    def create_start_and_end_timekeys(self):
        """Creates the start and end keys of the joint animation.

        TODO: describe
        """
        if self.get_time_key_by_time(0) is None:
            self.add_time_key(0, 0)
        if self.get_time_key_by_time(self.duration) is None:
            self.add_time_key(self.duration, 1)

   
    def compute_key_time(self, real_time):
        """Computes the corresponding key time for the given real time.

        The values between time keys are obtained by linear interpolation.

        Parameters:
        ----------
        real_time : float
            The real time where to compute the corresponding key time.

        Returns:
        ----------
        float
            The calculated key time.
        """
        self.create_start_and_end_timekeys()

        if self.get_time_key_by_time(real_time) is None:
            return_value = np.interp(real_time,
                [time_key[0] for time_key in self.time_keys],
                [time_key[1] for time_key in self.time_keys])
            return return_value
        else:
            return self.get_time_key_by_time(real_time)[1]
            
    def get_previous_key(self, key_time):
        """Returns the previous key of the joint animation at the given key time.

        Returns:
        ----------
        Transform
            The previous key.
        """
        
        previous_key = self.keys[0]

        for key in self.keys:
            if key.time < key_time:
               previous_key = key      
        return previous_key

    def get_next_key(self, key_time):
        """Returns the next key of the joint animation at the given key time.
        
        Returns:
        ----------
        Transform
          The next key.
        """

        next_key = self.keys[-1]

        for key in self.keys:
            if key.time > key_time:
                next_key = key
                break
        return next_key
    
    def get_lerp_transform(self, key_time):
        """
        Returns the linearly interpolated transform for the given key time.

        Parameters:
        ----------
        key_time : float
            The key time for which to compute the transform.
        
        Returns:
        ----------
        Transform
            The linearly interpolated transform.
        """
        previous_key = self.get_previous_key(key_time)
        next_key = self.get_next_key(key_time)
        delta = next_key.time - previous_key.time
        if math.isclose(delta, 0, rel_tol=1e-09, abs_tol=1e-10):
            return previous_key.transform
        else:
            progress = (key_time - previous_key.time) / delta
            return Transform.lerp(previous_key.transform, next_key.transform, progress)




            


                

        
