import numpy as np
import quaternion
import warnings

from skeletal_animation.core.math.constants import LOW_A_TOL, LOW_R_TOL

class Transform:
    """A class to represent a Transformation. 

    Note: np.quaternion is used to represent quaternions.
          Here are a few static values to declare quaternions: 
            quaternion.zero = np.quaternion(0, 0, 0, 0)
            quaternion.one = np.quaternion(1, 0, 0, 0) # identity quaternion
            quaternion.x = np.quaternion(0, 1, 0, 0)
            quaternion.y = np.quaternion(0, 0, 1, 0)
            quaternion.z = np.quaternion(0, 0, 0, 1)


    Attributes
    ----------
    pos : (float) ndarray
        1D array containing the 3 values for the position in this order [x, y, z].
    orient : np.quaternion
        Quaternion object defining the orientation. More info on np.quaternion here: https://quaternion.readthedocs.io/en/latest/#
    matrix : (float) ndarray
        4x4 matrix of the transformation corresponding to a translation by the position vector followed by a rotation by the quaternion.  

    """

    def __init__(self, pos = np.zeros(3), orient = quaternion.one):
        """**Default constructor for the Trasform class.**

        Parameters
        ----------
        pos : (float) ndarray, optional
            1D array containing the 3 values for the transform's position, by default [0, 0, 0].
        orient : np.quaternion, optional
            Quaternion object defining the transform's orientation, by default identity quaternion.
        matrix : (float) ndarray, optional
            If available, thef corresponding 4x4 transformation matrix, by default None.
        """
        self.pos = pos
        orient_norm = np.linalg.norm(quaternion.as_float_array(orient))
        if not np.isclose(orient_norm, 1, rtol=LOW_R_TOL, atol=LOW_A_TOL):
            warnings.warn("Transform built with non unit quaternion.", stacklevel=2)
        self.orient = orient

    @classmethod
    def from_matrix(cls, matrix=np.eye(4)):
        """Transform constructor from 4x4 transform matrix.

        Parameters
        ----------
        matrix : (float) ndarray), optional
            4x4 transformation matrix from which to extract position vector and orientation quaternion, by default 4x4 identity matri).

        Returns
        -------
        Transform
            Transform object.
        """
        return cls(matrix[3, 0:3], quaternion.from_rotation_matrix(matrix[0:3,0:3].transpose()))

    def as_matrix(self):
        """Computes matrix attribute. 

        Note: happens on object creation if no transform matrix is given as parameter.
        """
        t_matrix = np.eye(4)
        t_matrix[0:3, 0:3] = quaternion.as_rotation_matrix(self.orient).transpose()
        
        t_matrix[3, 0:3] = self.pos
        return t_matrix

    def normalize_quaternion(self):
        """Normalizes the quaternion of the transform.

        Returns
        -------
        Transform
            Transform with normalized quaternion.
        """
        return Transform(pos=self.pos, orient=self.orient.normalized())
    
    def inversed(self):
        """Returns the inverse of the transform.

        Returns
        -------
        Transform
            Inverse transform.
        """
        return Transform.from_matrix(np.linalg.inv(self.as_matrix()))

    def multiply_by(self, other):
        """Generates a Transform that is the result of the multiplication of two transforms.

        Note: uses matrix multiplication.

        Parameters
        ----------
        other : Transform
            Transform to multiply by.

        Returns
        -------
        Transform
            Resulting Transform. Corresponds to a transformation by the second matrix (other) followed by a transformation by the first matrix (self).
        """
        self_orient_norm = np.linalg.norm(quaternion.as_float_array(self.orient))
        other_orient_norm = np.linalg.norm(quaternion.as_float_array(other.orient))
        if not np.allclose([self_orient_norm, other_orient_norm], [1., 1.], rtol=LOW_R_TOL, atol=LOW_A_TOL):
            warnings.warn("Transform multiplication attempted with non unit quaternions. This can lead to erroneous results.", stacklevel=2)

        return Transform.from_matrix(np.matmul(self.as_matrix(), other.as_matrix()))

    def multiply_all(self, transforms):
        """Apply all transformations in the given list to the transform.
        
        Parameters
        ----------
        transforms : (Transform) list
            List of transforms to apply to the transform.

        Returns
        -------
        Transform
            Resulting transform.
        """
        #TODO
        pass        

    def translate(self, pos_vector):
        """Translates the transform by the given position vector.

        Parameters
        ----------
        pos : (float) ndarray, optional
            1D array containing the 3 values for the translation.

        Returns
        -------
        Transform
            Translated transform.
        """
        translation_transform = Transform(pos=pos_vector)
        return self.multiply_by(translation_transform)

    def rotate_quaternion(self, quaternion):
        """Rotates the transform by the given quaternion.

        Parameters
        ----------
        quaternion : np.quaternion
            Quaternion to rotate by.

        Returns
        -------
        Transform
            Rotated transform.
        """
        rotation_transform = Transform(orient=quaternion)
        return self.multiply_by(rotation_transform)   

    @staticmethod
    def lerp(t1, t2, progress):
        """Linear interpolation between two transforms.
            
            Parameters
            ----------
            t1 : Transform
                First transform.
            t2 : Transform
                Second transform.
            progress : float
                Progress between 0 and 1.        
        """
        px = np.interp(progress, [0, 1], [t1.pos[0], t2.pos[0]])
        py = np.interp(progress, [0, 1], [t1.pos[1], t2.pos[1]])
        pz = np.interp(progress, [0, 1], [t1.pos[2], t2.pos[2]])
        p = np.array([px, py, pz])
        return Transform(p, quaternion.slerp(t1.orient, t2.orient, 0, 1, progress))

    """
    # In case it's needed for a more intuitive rotation
    def rotate_euler(self, x, y, z):
        \"""Rotates the transform by the given Euler angles.

        Parameters
        ----------
        x : float
            Rotation around the x axis.
        y : float
            Rotation around the y axis.
        z : float
            Rotation around the z axis.

        Returns
        -------
        Transform
            Rotated transform.
        \"""
        rotation_transform = Transform(quaternion.from_euler_angles(x, y, z))
        return self.multiply_by(rotation_transform)
    """
    
    def __eq__(self, other):
        """Equality for Transform objects.

        Parameters
        ----------
        other : Transform
            Transform object to compare to.

        Returns
        -------
        Bool
            True if the 3 values for both positions are equal to each other, and both quaternions are equal. False otherwise.       
        """
        return (np.allclose(self.pos, other.pos, rtol=LOW_R_TOL, atol=LOW_A_TOL) and np.allclose(self.orient, other.orient, rtol=LOW_R_TOL, atol=LOW_A_TOL))

    def __matmul__(self, other):
        """Matrix multiplication for Transform objects.

        Parameters
        ----------
        other : Transform
            Transform object to multiply by.

        Returns
        -------
        Transform
            Resulting transform.
        """
        return self.multiply_by(other)

    def __repr__(self):
        return str(self.__dict__)