import numpy as np
import json
import warnings
from ..math.transform import Transform

class Joint:
    """A class to represent a joint (or bone).
    
    Attributes
    ----------
    LAST_ID : int
        Auto incrementing ID for the next join.
    id : int
        Unique id of the joint.
    name : str
        Name given to the joint. Not necessarily unique.    
    parent : Joint
        Unique parent Joint of the joint.
    children : (Joint) list
        List of children joints.
    local_bind_transform : Transform
        This is the initial transform of the joint in the parent joint's local coordinate system.
    inverse_model_bind_transform : Transform
        This is the inverse of the initial transform of the joint in the model's coordinate system.
    """
    LAST_ID = 0

    def __init__(self, name="new joint", local_bind_transform=Transform()):
        """**Default constructor for the Joint class.**

        TODO update docstrings

        Parameters
        ----------
        name : str, optional
            Name of the joint, by default "new joint".
        local_bind_transform : _type_, optional
            Initial transform relative to the parent, by default identity matrix.
        """
        self.id = Joint.LAST_ID
        Joint.LAST_ID += 1
        self.name = name
        self.parent = None   
        self.children = []
        self.local_bind_transform = local_bind_transform
        self.current_model_transform = None 
        
    def add_child(self, joint):
        """Makes the joint a child of the current joint.

        Note: child joint's parent is updated accordingly.

        Parameters
        ----------
        joint : Joint
            The joint to be added as a child.

        Warnings
        --------
        If the addition of the joint makes the tree cyclic, a warning is raised.        
        """
        if joint.parent is None:
            joint.parent = self
            self.children.append(joint)
        elif joint.parent != self:
            old_parent = joint.parent
            old_parent.children.remove(joint)
            joint.parent = self
            self.children.append(joint)
        if self.is_cyclic():              
           warnings.warn("Cyclic joint tree detected.", stacklevel=2)
   
    def add_children(self, joints):
        """Makes the joints children of the current joint.

        Parameters
        ----------
        joints : (Joint) list
            The joints to be added as children.
        """
        for joint in joints:
            self.add_child(joint)

    def remove_child(self, joint):
        """Removes the joint from the children of the current joint.

        Note: If the joint is not a child of the current joint, then nothing happens.
        Note: Child joint's parent is updated accordingly.
        
        Parameters
        ----------
        joint : Joint
            The joint to be removed.
        """
        if joint in self.children:
            self.children.remove(joint)
            joint.parent = None

    def remove_children(self, joints):
        """Removes the joints from the children of the current joint.

        Parameters
        ----------
        joints : (Joint) list
            The joints to be removed.
        """
        for joint in joints:
            self.remove_child(joint)

    def remove_all_children(self):
        """Removes all the children of the current joint.
        """
        for joint in self.children:
            self.remove_child(joint)

    # def compute_inverse_model_bind_transform(self, parent_model_bind_transform_matrix):
    #     """Computes the inverse model bind transform for the entire joint hierarchy.
        
    #     Note: this method is recursive.
    #     """
    #     model_bind_transform_matrix = np.matmul(parent_model_bind_transform_matrix, self.local_bind_transform.as_matrix())
    #     self.inverse_model_bind_transform = Transform.from_matrix(np.linalg.inv(model_bind_transform_matrix))
    #     for child_joint in self.children:
    #         child_joint.compute_inverse_model_bind_transform(model_bind_transform_matrix)
    
    def forward_kinematics(self, current_pose_dict, parent_model_transform):
        """Computes the animated transform for the joint.

        Parameters
        ----------
        current_pose : Pose
            The pose dict for the current frame.
        parent_animated_transform_matrix : numpy.ndarray
            The animated transform of the parent joint.
        """
        if current_pose_dict == None:
            current_local_transform = Transform()
        else:
            current_local_transform = current_pose_dict[self.id]
        current_model_transform = current_local_transform @ self.local_bind_transform @ parent_model_transform
        self.current_model_transform = current_model_transform        
        for child_joint in self.children:
            child_joint.forward_kinematics(current_pose_dict, current_model_transform)        
        
    def __repr__(self):
        dict = {}
        dict["id"] = self.id
        dict["name"] = self.name
        if self.parent is None:
            dict["parent"] = None
        else:
            dict["parent"] = self.parent.name
        dict["children"] = [child.name for child in self.children]
        dict["local_bind_transform"] = self.local_bind_transform.__repr__()
        dict["inverse_model_bind_transform"] = self.inverse_model_bind_transform.__repr__()
        dict["animated_transform"] = self.animated_transform.__repr__()
        return json.dumps(dict, indent=4)

    def is_cyclic(self):
        """Checks for cyclicity in the joint hierarchy for the current joint.

        Note: The joint hierarchy is considered cyclic only if a joint is the parent of one of its ancestors.

        Examples
        -------
        >>> root = Joint()
        >>> child1 = Joint()
        >>> child2 = Joint()
        >>> child3 = Joint()
        >>> root.add_children([child1, child2])
        >>> child1.add_child(child3)
        # The joint hierarchy is now:
        #   root -> child1 -> child3
        #        -> child2
        >>> root.is_cyclic()
        False
        >>> child3.add_child(child1)
        # The joint hierarchies are now:
        # root -> child2  and  ... -> child1 -> child3 -> child1 -> child3 -> child1 -> ...
        >>> root.is_cyclic()
        False
        >>> child1.is_cyclic()
        True
        
        Returns
        -------
        bool
            True if the joint hierarchy is cyclic, False otherwise.
        """

        # List of visited nodes.
        visited = []
        # Stack to keep track of the current node.
        stack = []
        stack.append(self)

        # Run till stack is empty or cycle is found.
        while stack:
            node = stack.pop()
            for child in node.children:
                if child in visited:
                    return True
                else:
                    stack.append(child)
                    visited.append(node)

        # If we reach here, then there is no cycle
        return False