import numpy as np
from .joint import Joint

class Skeleton:
    """A class to represent a skeleton (or rig).

    Attributes
    ----------
    name : str
        Name given to the skeleton. Not necessarily unique.
    root : Joint
        Root joint of the hierarchy for the skeleton.    
    """

    def __init__(self, root=None, name="new skeleton"):
        """**Default constructor for the Skeleton class.**
        
        Parameters
        ----------
        root : Joint, optional
            Root joint of the hierarchy for the skeleton, by default None.
        name : str, optional
            Name of the skeleton, by default "new skeleton".        
        """
        self.name = name
        self.root = root
    
    def as_joint_list(self):
        """Returns the skeleton as a list of joints.
        
        Returns
        -------
        (Joint) list
            List of joints in the skeleton.
        """

        # Convert a directed graph into an adjacency list
        adjacency_list = []

        def dfs(joint):
            adjacency_list.append(joint)
            for child in joint.children:
                dfs(child)

        dfs(self.root)
        return adjacency_list

    def compute_bind_pose(self):
        """Computes the inverse model bind transform for each joint in the skeleton's joint hierarchy.
        """
        self.root.forward_kinematics(None, self.root.local_bind_transform)

    def compute_pose(self, pose_dict):
        """Computes the animated transform for each joint in the skeleton's joint hierarchy.
        
        Parameters
        ----------
        pose_dict : Dict[int, Transform]
        """
        self.root.forward_kinematics(pose_dict, self.root.local_bind_transform)


        
