from logging import root
import unittest
import numpy as np
import quaternion

from skeletal_animation.core.math.transform import Transform
from skeletal_animation.core.math.constants import HIGH_R_TOL, HIGH_A_TOL
from skeletal_animation.core.model.joint import Joint

class TestJointDefaultConstructor(unittest.TestCase):
    def setUp(self): 
        self.root = Joint(name="root")
        self.back = Joint(name="back", local_bind_transform=Transform(pos=[1,0,1]))
        self.root.add_child(self.back)
        self.arm = Joint(name="arm", local_bind_transform=Transform(pos=[0.25,0,0], orient=np.quaternion(1,0,0,0)))
        self.back.add_child(self.arm)
        with self.assertWarns(UserWarning):
            self.hand = Joint(name="hand", local_bind_transform=Transform(orient=quaternion.from_float_array([.2,.8,.2,0])))
        self.arm.add_child(self.hand)
        # root -> back -> arm -> hand
        self.root.compute_inverse_model_bind_transform(self.root.local_bind_transform.as_matrix())
        
    def test_default_constructor_with_default_parameters(self):
        j = Joint()
        self.assertIsNone(j.parent)
        self.assertEqual(j.name, "new joint")
        self.assertEqual(j.local_bind_transform, Transform())

    def test_default_constructor_with_no_parent(self):
        t = Transform([1,0,2], np.quaternion(0, 2/3, 1/3, 2/3))
        ti = Transform.from_matrix(np.linalg.inv(t.as_matrix()))
        j = Joint(name="root", local_bind_transform=t)
        self.assertIsNone(j.parent)
        self.assertEqual(j.name, "root")
        self.assertEqual(j.local_bind_transform, t)
    
    def test_default_constructor_with_root_parent(self):          
        j = self.back
        self.assertEqual(j.parent, self.root)
        self.assertEqual(j.inverse_model_bind_transform, j.local_bind_transform.inversed())
    
    # TODO Fix inverse bind transform calc
    def test_default_constructor_with_non_root_parent(self):  
        j = self.hand
        self.assertEqual(j.parent.parent.parent, self.root)
        print(j.inverse_model_bind_transform.as_matrix())
        print(np.matmul(
                self.arm.children[0].local_bind_transform.as_matrix(),
                self.arm.inverse_model_bind_transform.as_matrix(),
                ))        

        self.assertEqual(j.inverse_model_bind_transform,
            Transform.from_matrix(np.matmul(
                self.arm.inverse_model_bind_transform.as_matrix(),
                self.arm.children[0].local_bind_transform.as_matrix()
                ))
            )

class TestJointIsCyclic(unittest.TestCase):
    def setUp(self): 
        self.root = Joint(name="root")
        self.back = Joint(name="back", local_bind_transform=Transform(pos=[1,0,1]))
        self.root.add_child(self.back)
        # root -> back
    
    def test_is_cyclic_simple(self):
        self.assertFalse(self.root.is_cyclic())
        with self.assertWarns(UserWarning):
            self.back.add_child(self.root)
        # root -> back -> root  =>  CYCLE!
        self.assertTrue(self.root.is_cyclic())
        self.back.remove_child(self.root)
        self.assertFalse(self.root.is_cyclic())

    def test_is_cyclic_intermediate(self):
        r = Joint(name="r")
        c1 = Joint(name="c1")
        c2 = Joint(name="c2")
        c3 = Joint(name="c3")
        r.add_children([c1, c2])
        c1.add_child(c3)
        self.assertFalse(r.is_cyclic())
        with self.assertWarns(UserWarning):
            c3.add_child(c1)
        self.assertFalse(r.is_cyclic())
        self.assertTrue(c1.is_cyclic())
        c3.remove_child(c1)
        self.assertFalse(c1.is_cyclic())
        c3.add_child(r)
        self.assertFalse(c1.is_cyclic())

    
    def test_is_cyclic_complex(self):
        arm = Joint(name="arm", local_bind_transform=Transform())
        self.back.add_child(arm)
        forearm = Joint(name="forearm", local_bind_transform=Transform())
        arm.add_child(forearm)
        hand = Joint(name="hand", local_bind_transform=Transform())
        forearm.add_child(hand)
        finger = Joint(name="finger", local_bind_transform=Transform())
        hand.add_child(finger)
        fingertip = Joint(name="fingertip", local_bind_transform=Transform())
        finger.add_child(fingertip)
        head = Joint(name="head", local_bind_transform=Transform())
        self.back.add_child(head)
        nose = Joint(name="nose", local_bind_transform=Transform())
        head.add_child(nose)
        # root -> back -> arm -> forearm -> hand -> finger -> fingertip
        #              -> head -> nose
        
        self.assertFalse(self.root.is_cyclic())
        self.assertFalse(arm.is_cyclic())
        with self.assertWarns(UserWarning):
            finger.add_child(self.back)
        # root -> 
        # back -> arm -> forearm -> hand -> finger -> fingertip
        #      -> head -> nose -> arm              -> back -> arm -> forearm -> hand -> finger -> fingertip
        #                                                  -> head -> nose -> arm 
        # => CYCLE!  
        self.assertFalse(self.root.is_cyclic())      
        self.assertTrue(arm.is_cyclic())


class TestJointAddChild(unittest.TestCase):
    def setUp(self): 
        self.root = Joint(name="root")
        self.back = Joint(name="back")
        self.arm = Joint(name="arm")
        # root -> back -> arm
    
    def test_parent_update(self):
        self.assertIsNone(self.back.parent)
        self.root.add_child(self.back)
        self.assertEqual(self.back.parent, self.root)
        self.assertEqual(self.root.children, [self.back])
    
    def test_cyclicity_warning(self):
        self.root.add_child(self.back)
        self.back.add_child(self.arm)
        with self.assertWarns(UserWarning):
            self.arm.add_child(self.root)

class TestJointRemoveChild(unittest.TestCase):
    def setUp(self): 
        self.root = Joint(name="root")
        self.back = Joint(name="back")
        self.arm = Joint(name="arm")
        # root -> back -> arm

    def test_parent_update(self):
        self.root.add_child(self.back)
        self.back.add_child(self.arm)
        self.root.remove_child(self.back)
        self.assertIsNone(self.back.parent)
        self.assertEqual(self.root.children, [])
    
class TestJointComputeInverseModelBindTransform(unittest.TestCase):
    def setUp(self):
        self.root = Joint(name="root")
        self.back = Joint(name="back", local_bind_transform=Transform(pos=[1,0,1]))
        self.root.add_child(self.back)
        self.arm = Joint(name="arm", local_bind_transform=Transform(pos=[1,2,3], orient=np.quaternion(0,2/3,1/3,2/3)))
        self.back.add_child(self.arm)

    # TODO fix this test or fix the class
    def test_compute_inverse_model_bind_transform(self):
        mi = [[-1/9, 4/9, -8/9, 0],
                [-4/9, 7/9, 4/9, 0],
                [8/9, 4/9, 1/9, 0],
                [-5/3, -10/3, -1/3, 1]] # inverse of self.arm.local_bind_transform      
        self.assertIsNone(self.arm.inverse_model_bind_transform)
        self.root.compute_inverse_model_bind_transform(self.root.local_bind_transform.as_matrix())

        #print(self.arm.inverse_model_bind_transform.as_matrix())
        #print(self.back.inverse_model_bind_transform.inversed().multiply_by(self.arm.local_bind_transform).as_matrix())
        self.assertTrue(np.allclose(self.arm.inverse_model_bind_transform.as_matrix(),
            np.matmul(self.back.inverse_model_bind_transform.inversed().as_matrix(), self.arm.local_bind_transform.as_matrix()), rtol=HIGH_R_TOL, atol=HIGH_A_TOL))
        
